package JNSA::DB;	# -*-Perl-*-

use DBI;
use Data::Dumper;
use vars qw($VERSION);

$VERSION = '1.91';

sub new {
    my $class = shift;
    my $attr = shift;
    my $self = {
	dbh => undef,
	config => undef,
	sql => {
	    query => [],
	    exec => []
	},
	error => {
	    connect => undef,
	    query => [],
	    exec => [],
	    disconnect => undef,
	},
	errorMessage => '',
	resultSetArray => [],
	state => {
	    isConnected => 0,
	},
    };
    _setAttr($self,$attr);
    bless $self, $class;
    return $self;
}

sub _setAttr {
    my ($self,$attr) = @_;
    my $key;
    for $key (keys(%{$attr})) {$self->{$key} = $attr->{$key};}
}

sub setConfig {
    my $self = shift;
    $self->{config} = shift;
}

#==========================================================================
# connect/disconnect
#==========================================================================
sub connect {
    my $self = shift;

    eval {
	my $cfg = $self->{config};
	# AutoCommit=>0 ϳ
	$self->{dbh} = DBI->connect($cfg->{dsn},$cfg->{usr},$cfg->{pw},
			    { RaiseError=>1, PrintError=>0 });
    };
    $self->{error}->{connect} = $@;
    $self->{state}->{isConnected} = 1 if ($@ eq '');
}

sub disconnect {
    my $self = shift;

    eval {
	$self->{dbh}->disconnect if ($self->{state}->{isConnected});
    };
    $self->{error}->{disconnect} = $@;
    $self->{state}->{isConnected} = 0;
}

#==========================================================================
# rollback
#==========================================================================


#==========================================================================
# execute
#==========================================================================
sub execSession {
    my ($self,$sqls) = @_;

    $self->connect;
    return if ! $self->{state}->{isConnected};

    $self->{dbh}->{AutoCommit} = 0;
    $self->execSqlArray($sqls);
    if ($self->isNoExecError) {
	#print STDERR Dumper($self);
	$self->{dbh}->commit;
    } else {
	$self->{dbh}->rollback;
    }
    $self->disconnect;
    $self->{errorMessage} = $self->getErrorMessageString;
}

sub execSqlArray {
    my ($self,$sqls) = @_;
    my $i;

    for $i (0..$#{$sqls}) {
	$self->execSql($sqls->[$i]);
    }
}

sub execSql {
    my ($self,$sql) = @_;
    my $dbh = $self->{dbh};

    push @{$self->{sql}->{exec}}, $sql;

    $@ = "";
    if ($self->{state}->{isConnected}) {
	eval {
	    my $sth = $dbh->prepare($sql);
	    $sth->execute;
	    $sth->finish;
	};
    }
    push @{$self->{error}->{exec}}, $@;
}


#==========================================================================
# query
#==========================================================================
sub querySession {
    my ($self,$sqls) = @_;

    $self->connect;
    $self->querySqlArray($sqls);
    $self->disconnect;
    $self->{errorMessage} = $self->getErrorMessageString;
}

sub querySqlArray {
    my ($self,$sqls) = @_;
    my $i;

    for $i (0..$#{$sqls}) {
	my $t;
	$self->querySqlToArray($sqls->[$i],$t);
    }
}

sub querySqlToArray {
    my ($self,$sql,$a) = @_;
    my $dbh = $self->{dbh};
    my @row;

    push @{$self->{sql}->{query}}, $sql;

    $a = [];
    $@ = "";
    if ($self->{state}->{isConnected}) {
	eval {
	    #$a = $dbh->selectall_arrayref($sql);

	    my $sth = $dbh->prepare($sql);
	    $sth->execute;

	    my $n = 0;
	    my $rowref;
	    while (@row = $sth->fetchrow_array) {
		$a->[$n] = [@row];
		$n++;
	    }
	    $sth->finish;
	};
    }
    push @{$self->{resultSetArray}}, $a;
    push @{$self->{error}->{query}}, $@;
}

#==========================================================================
# getMax
#==========================================================================
sub getMax {
    my ($self,$item,$tbl) = @_;
    my $a = [];

    $self->querySession(["SELECT max($item) FROM $tbl;"]);

    return $self->{resultSetArray}->[0]->[0]->[0];
}

#==========================================================================
# recordExists
#==========================================================================
sub recordExists {
    my ($self,$tbl,$item,$value) = @_;
    my $a = [];

    $self->querySession(["SELECT $item FROM $tbl WHERE $item = $value;"]);

    return ($#{$self->{resultSetArray}->[$#{$self->{resultSetArray}}]} < 0) ? 0 : 1;
}

#==========================================================================
# ErrorMessage
#==========================================================================

sub getErrorMessageString {
    my $self = shift;
    my $error = $self->{error};
    my $i;
    my $msg;
    my $s = "";

    $msg = $error->{connect};
    $msg =~ s/ at \S+ line \d+//;
    if ($msg ne '') {
	$s .= "== CONNECT ========================\n";
	$s .= "$msg\n";
    }

    for $i (0..$#{$error->{query}}) {
	$msg = $error->{query}->[$i];
	$msg =~ s/ at \S+ line \d+//;
	if ($msg ne '') {
	    my $ii = $i + 1;
	    $s .= "== QUERY $ii ========================\n";
	    $s .= "SQL:" . $self->{sql}->{query}->[$i] . "\n";
	    $s .= "MSG:" . $msg . "\n";
	}
    }

    for $i (0..$#{$error->{exec}}) {
	$msg = $error->{exec}->[$i];
	$msg =~ s/ at \S+ line \d+//;
	if ($msg ne '') {
	    my $ii = $i + 1;
	    $s .= "== EXEC $ii ========================\n";
	    $s .= "SQL:" . $self->{sql}->{exec}->[$i] . "\n";
	    $s .= "MSG:" . $msg . "\n";
	}
    }

    $msg = $error->{disconnect};
    $msg =~ s/ at \S+ line \d+//;
    if ($msg ne '') {
	$s .= "== DISCONNECT ========================\n";
	$s .= "$msg\n";
    }
    return $s;
}


#==========================================================================
# print
#==========================================================================
sub printErrorIfError {
    my $self = shift;
    $self->printError if $self->isError;
}

sub printError {
    my $self = shift;
    $self->{errorMessage} = $self->getErrorMessageString;
    print $self->{errorMessage};
}

sub printErrorAssumption {
    my($self,$errorMessage) = @_;
    my $s = ""; 

    $_ = $errorMessage;
    
    if (/parse error at or near \"/) {
	$s .= "
- some field which must be specified are empty.
- bug because of some single quotes.
";
    }
    if (/Attribute '.+' not found\./) {
	$s .= "
- wrong values are specified to some numeric or boolean fields.
";
    }
    if (/Cannot insert a duplicate key/) {
	$s .= "
- A primary key constraint violation.
";
    }
    if (/Database \".+\" does not exist in the system catalog/) {
	$s .= "
- The database is not exists.
";
    }
    if (/Connection refused/) {
	$s .= "
- Its database connection was refused.
";
    }
    if (/install_driver\(.+\) failed:/) {
	$s .= "
- Perl DBD specified in configuration file is not found.
";
    }
    if (/no database driver/) {
	$s .= "
- A wrong Perl database driver is specified.
";
    }
    if (/ still referenced from / ||
	/ referential integrity violation /) {
	$s .= "
- this record can't be deleted because of referential integrity. First you must delete some records which refer to this record.
";
    }
    if (/ value too long for type character /) {
	$s .= "
- the number of characters in some string are exceeded.
";
    }
    if (/ Unterminated quoted identifier/) {
	$s .= "
- misplaced single quote is entered.
";
    }
    if (/ overflow on numeric ABS/) {
	$s .= "
- some numbers are out of range.
";
    }
    if (/  ([^:]+): permission denied/) {
	$s .= "Database table '$1' permission denied.\n";
    }

    if ($s ne "") {
	print "<PRE>The database update was failed as following reasons.\n$s</PRE>\n";
    } else {
	print "<PRE>Unknown error. See the log in below.\n$_\n</PRE>\n";
    }
}

#==========================================================================
# error checking
#==========================================================================
sub isError {
    my $self = shift;
    my $k;

    for $k (keys(%{$self->{error}})) {
	my $value = $self->{error}->{$k};
	my $ref = ref($value);
	return 1 if ($ref eq '' && $value ne '');
	return 1 if ($ref eq 'ARRAY' && _isErrorArray($value));
    }
    return 0;
}

sub isNoExecError  {return (! _isErrorArray($_[0]->{error}->{exec}));}
sub isNoQueryError {return (! _isErrorArray($_[0]->{error}->{query}));}

sub _isErrorArray {
    my $v = shift;
    my $i;
    for $i (0..$#{$v}) {return 1 if ($v->[$i] ne '');}
    return 0;
}

1;
