/* $Id: genkeyp.c,v 1.2 2004/07/09 09:48:37 urushima Exp $
 *******************************************************************
 *
 * Challenge PKI Test Suite 2.0
 *
 *							Jul 7, 2004
 *
 *			Copyright (c) 2002-2004 
 *			Challenge PKI Project http://www.jnsa.org/mpki/
 *			  IPA-ISEC http://www.ipa.go.jp/security/
 *			  NPO-JNSA http://www.jnsa.org/
 *			  SECOM Co.,Ltd. http://www.secom.co.jp/
 *			  All Rights Reserved.
 *
 *******************************************************************
 */

/*
 * genkeyp.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ok_x509.h"
#include "ok_asn1.h"
#include "ok_sha1.h"
#include "ok_pem.h"

void dbgmsg(char *func, char *msg);
void fprint_sha1_160bit_keyid(FILE *fp, unsigned char *pHash);
void fprint_sha1_64bit_keyid(FILE *fp, unsigned char *pHash);

/* options */
char PRIV_KEY_FILE[256] = "/tmp/priv.pem";
char PUB_KEY_FILE[256]  = "/tmp/pub.pem";
int  KEY_SIZE           = 512;
char KEY_TYPE[256]      = "rsa";
char KEY_ID_FILE[256]   = "/tmp/keyid.txt";


void usage() {
    printf("usage: genkeyp\n");
    printf("\t-type type : rsa, dsa [default:rsa]\n");
    printf("\t-size  num : key size [default:512bit]\n");
    printf("\t-priv file : output file for private key [default:/tmp/priv.pem]\n");
    printf("\t-pub  file : output file for public key [default:/tmp/pub.pem]\n");
    printf("\t-kid  file : output file for KeyIdentifier [default:/tmp/keyid.txt]\n");
    printf("\t-not160kid : KeyIdentifier [default:SHA1 160bit]\n");
}


/* options
 *
 */
void options(int argc, char **argv) {
    int i;

    for(i=1; i<argc; i++) {
        if(!strcmp("-type", argv[i])) {
            if (++i < argc) strcpy(KEY_TYPE, argv[i]);
            else goto error;
        } else if(!strcmp("-size", argv[i])) {
            if (++i < argc) KEY_SIZE = atoi(argv[i]);
            else goto error;
        } else if(!strcmp("-priv", argv[i])) {
	    if (++i < argc) strcpy(PRIV_KEY_FILE, argv[i]);
            else goto error;
        } else if(!strcmp("-pub", argv[i])) {
            if (++i < argc) strcpy(PUB_KEY_FILE, argv[i]);
            else goto error;
        } else if(!strcmp("-kid", argv[i])) {
            if (++i < argc) strcpy(KEY_ID_FILE, argv[i]);
            else goto error;
        } else {
            goto error;
        }
    }

    return;

error:
    usage();
    exit(-1);
}


void genkeyid(unsigned char *der, int sz) {

    FILE *fp;
    int i, j;
    unsigned char hash[32], buf[32];

#ifdef DEBUG
    printf("SubjectPublicKeyInfo, size = %d\n", sz);
    for (i=0; i<sz; i++) {
        printf("[%02x]", *(der+i));
    }
#endif

    if ((fp = fopen(KEY_ID_FILE, "w")) == NULL) {
        printf("error: fopen() for %s\n", KEY_ID_FILE);
        exit(-1);
    }

    /*
     * The format of paramter "der" is below:
     * SubjectPublicKeyInfo  ::=  SEQUENCE {
     *    algorithm            AlgorithmIdentifier,
     *    subjectPublicKey     BIT STRING }
     */

    /* Finally "j" is to point the top of BIT STRING of subjectPublicKey
     * excluding the tag, length.
     */
    j = 0; 
	      
    /* the top SEUQUENCE  */
    j++;
    if (*(der+j) & 0x80) {
        j += 1 + (*(der+j) & 0x7f);
    } else {
        j += 1;
    }

    /* SEUQUENCE for algorithm */
    j++;
    j += 1 + *(der+j);

    /* subjectPublicKey */
    if (*(der+j) != 0x03) {
        printf("Unexpected format error in genkeyid().");
	exit;
    }
    /* Get the top of the BIT STRING excluding the tag, length */
    j++;
    if (*(der+j) & 0x80) {
        j += 1 + (*(der+j) & 0x7f);
    } else {
        j += 1;
    }
    /* and excluding unused bits */
    j += 1; 

#ifdef DEBUG
    printf("\nTo be hashed (BIT STRING excluding the tag, length, unused bits)\n");
    for (i=0; i<sz-j; i++) {
        printf("[%02x]", *(der+j+i));
    }
    /*
     * if "der + j" is correct pointer, "mogemo.der" will be
     * SEQUENCE { INTEGER ... INTEGER ... }
     */
    ASN1_write_der(der + j, "mogemo.der");
#endif

    memset(buf, 0x00, sizeof(hash));
    OK_SHA1(sz-j, der+j, hash);

#ifdef DEBUG
    printf("\nSHA1 bits(160bits)\n");
    for (i=0; i<20; i++) {
        printf("[%02x]", hash[i]);
    }
#endif

    fprint_sha1_160bit_keyid(fp, &hash[0]);
    fprint_sha1_64bit_keyid(fp, &hash[0]);

    fclose(fp);
}

/* dsaKeyGen()
 *
 */
void dsaKeyGen() {

    char func[] = "dsaKeyGen()";
    Key *pub, *prv;
    DSAParam *pm;
    unsigned char *der = NULL;
    int i;

#ifdef DEBUG
dbgmsg(func, "DSA keygen start.");
#endif

    if ((KEY_SIZE<512) || (KEY_SIZE>1024)) {
        printf("error: bad key size ...\n");
        return;
    }
    if ((prv=(Key*)DSAprvkey_new()) == NULL) {
        printf("error: DSAprvkey_new()\n"); 
        exit(-1);
    }
    if ((pub=(Key*)DSApubkey_new()) == NULL) {
        printf("error: DSApubkey_new()\n"); 
        exit(-1);
    }
    if ((pm = DSAPm_gen_parameter(KEY_SIZE))==NULL) {
        printf("error: DSAPm_gen_parameter()\n"); 
        exit(-1);
    }
    if (DSAprv_generate(pm,(Prvkey_DSA*)prv)) {
        printf("error: DSAprv_generate()\n"); 
        exit(-1);
    }
    if (PEM_write_p8(prv, PRIV_KEY_FILE) < 0) {
        printf("error: PEM_write_p8() for private key\n"); 
        exit(-1);
    }

    /* public key 
     */
    DSAprv_2pub((Prvkey_DSA*)prv, (Pubkey_DSA*)pub);
    i = DSApub_estimate_der_size((Pubkey_DSA *)pub);
    if((der=(unsigned char *)calloc(i,sizeof(char))) == NULL){
      printf("error:calloc\n");
      exit(-1);
    }
    x509_DER_pubkey((Key *)pub,der,&i);
    pem_write(der,PUB_KEY_FILE,
	      "-----BEGIN PUBLIC KEY-----\n",
	      "-----END PUBLIC KEY-----\n");

    genkeyid(der, i);

    FREE(der); der = NULL;
    DSAkey_free(prv); DSAkey_free(pub); 
}


/* rsaKeyGen()
 *
 */
void rsaKeyGen() {

    char func[] = "rsaKeyGen()";
    Pubkey_RSA *pub;
    Prvkey_RSA *prv;

    int i, j;
    unsigned char *der = NULL, hash[32];

#ifdef DEBUG
dbgmsg(func, "RSA keygen start.");
#endif


    if ((prv = RSAprvkey_new()) == NULL) {
        printf("error: RSAprvkey_new()\n"); 
        exit(-1);
    }
    if ((pub = RSApubkey_new()) == NULL) {
        printf("error: RSApubkey_new()\n"); 
        exit(-1);
    }
    if (RSAprv_generate(prv, KEY_SIZE>>4)) {
        printf("error: RSAprv_generate()\n"); 
        exit(-1);
    }
    if (PEM_write_p8(prv, PRIV_KEY_FILE) < 0) {
        printf("error: PEM_write_p8() for private key\n"); 
        exit(-1);
    }


    /* public key 
     */
    RSAprv_2pub(prv, pub);
    if ((der = RSApub_toDER(pub, NULL, &j)) == NULL) {
        printf("error: RSApub_toDER()\n");
        exit(-1);
    }
    x509_DER_pubkey((Key *)pub,der,&j);
    pem_write(der,PUB_KEY_FILE,
	      "-----BEGIN PUBLIC KEY-----\n",
	      "-----END PUBLIC KEY-----\n");

    genkeyid(der, j);

    FREE(der); der = NULL;
    RSAkey_free((Key*)prv); RSAkey_free((Key*)pub);
}



#ifdef DEBUG
/* dbgmsg
 *
 */
void dbgmsg(char *func, char *msg) {
    printf("#DEBUG# %-16s : %s\n", func, msg);
}
#endif


/* main
 *
 */
int main(int argc, char **argv) {

    options(argc, argv);
#ifdef DEBUG
printf("KEY_TYPE[%s] KEY_SIZE[%d]\n", KEY_TYPE, KEY_SIZE);
printf("PRIV_KEY_FILE[%s] PUB_KEY_FILE[%s]\n", PRIV_KEY_FILE, PUB_KEY_FILE);
printf("KEY_ID_FILE[%s]\n", KEY_ID_FILE);
#endif    

    if (!strcmp(KEY_TYPE, "rsa")) {
        rsaKeyGen();
    } else if (!strcmp(KEY_TYPE, "dsa")) {
        dsaKeyGen();
    } else {
        printf("error: unkown key type\n");
    }	

    printf("\n");
    return 0;
}

void fprint_sha1_160bit_keyid(FILE *fp, unsigned char *pHash) {
  int	i;
  for (i = 0; i < 20; i++) {
    fprintf(fp, "%02x", pHash[i]);
  }
  fprintf(fp, "\n");
}

/*
 * fprint_sha1_64bit_keyid(fp, pHash)
 *
 *	0100-bits followed by the least significant 60 bits.
 */
void fprint_sha1_64bit_keyid(FILE *fp, unsigned char *pHash) {
  unsigned char buf[8];
  int	i;

  memcpy(buf, &pHash[12], 8);
  buf[0] = (buf[0] & 0x0f) | 0x40;

  for (i = 0; i < 8; i++) {
    fprintf(fp, "%02x", buf[i]);
  }
  fprintf(fp, "\n");
}

