/* $Id: ocspreq.c,v 1.4 2004/07/09 09:48:37 urushima Exp $
 *******************************************************************
 *
 * Challenge PKI Test Suite 2.0
 *
 *							Jul 7, 2004
 *
 *			Copyright (c) 2002-2004 
 *			Challenge PKI Project http://www.jnsa.org/mpki/
 *			  IPA-ISEC http://www.ipa.go.jp/security/
 *			  NPO-JNSA http://www.jnsa.org/
 *			  SECOM Co.,Ltd. http://www.secom.co.jp/
 *			  All Rights Reserved.
 *
 * ChangeLog
 *	1.0 -> 2.0
 *	result string was drastically changed.
 *	[11][22][33][44][FF] => 1212AA:11223344FF
 *	^nonce                  ^sn    ^nonce
 *
 *******************************************************************
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include"ok_asn1.h"

#define OID_NONCE    "1.3.6.1.5.5.7.48.1.2"

char errmsg[128];
int nonceLen = 0;

int OCSP_REQ_get_nonce(unsigned char *in, unsigned char *nonce, int sz);
int OCSP_REQ_get_serial(unsigned char *derReq, unsigned char **serial, int *sz);
int getNonce(unsigned char *in, char *out, int sz);
void dbgmsg(char *func, char *msg);

/*
 * main()
 *
 *	argv[0]: file name whose content is OCSP request
 *
 */
int main(int argc, char **argv) {
  unsigned char *reqder;	/* OCSP request in DER */
  unsigned char nonce[64];
  unsigned char *serial;
  int len_serial;

  int i;

  /* check the parameters */
  if (argc < 2) {
    printf("Usage: ocspreq filename\n");
    exit(-1);
  }

  if ((reqder=ASN1_read_der(argv[1])) == NULL) {
    printf("error in ASN1_read_der()");
    exit(-1);
  }

  if (OCSP_REQ_get_serial(reqder, &serial, &len_serial) == -1) {
    printf("error: %s\n", errmsg);
    exit(-1);
  }

  if (OCSP_REQ_get_nonce(reqder, nonce, sizeof(nonce)) == -1) {
    printf("error: %s\n", errmsg);
    exit(-1);
  }


  for (i = 0; i < len_serial; i++) printf("%02x", *(serial + i));
  
  printf(":");

  for (i = 0; i < nonceLen; i++) printf("%02x", nonce[i]);

  printf("\n");

  return 0;
}

/*
 * OCSP_REQ_get_serial
 *
 *	-derReq : 	string of DER data
 *	+serial :	beginning of the value field which is serialNumber.
 *	+sz :		the number of the bytes of the value field.
 */
int OCSP_REQ_get_serial(unsigned char *derReq, unsigned char **serial, int *sz) {
  unsigned char *derSeqLevel2;
  unsigned char *derSeqLevel3;
  unsigned char *derSeqLevel4;
  unsigned char *derSeqLevel5;
  unsigned char *derHashAlg;
  unsigned char *derIssuerNameHash;
  unsigned char *derIssuerKeyHash;
  unsigned char *derSerialNumber;

  if (derReq == NULL || *derReq != 0x30) {
    printf("ocspreq: malformed ocsp request\n");
    exit(-1);
  }

  derSeqLevel2 = ASN1_next(derReq);
  derSeqLevel3 = ASN1_next(derSeqLevel2);
  derSeqLevel4 = ASN1_next(derSeqLevel3);
  derSeqLevel5 = ASN1_next(derSeqLevel4);
  derHashAlg = ASN1_next(derSeqLevel5);

  derIssuerNameHash = ASN1_skip(derHashAlg);
  derIssuerKeyHash = ASN1_skip(derIssuerNameHash);
  derSerialNumber = ASN1_skip(derIssuerKeyHash);

  if (*derSerialNumber != ASN1_INTEGER) {
    printf("ERROR: ocspreq: malformed format. serialNumber not found.\n");
    exit(-1);
  }

  if (*(derSerialNumber + 1) & 128) {
    printf("ERROR: ocspreq: multi byte length tag not supported\n");
    printf("ERROR: ocspreq: serialNumber up to 20 octets by RFC 3280.\n");
    exit(-1);
  }

  *serial = derSerialNumber + 2;
  *sz = 127 & *(derSerialNumber + 1);

  return 0;
}

/*
 * OCSP_REQ_get_nonce
 *
 * Get Nonce value from the data pointed by a parameter.
 *
 * params in: pointer at ASN.1 in DER
 *        nonce: area where Nonce is to be set
 *        sz: size of nonce area
 *
 * return -1: some error happened
 *        not -1: success
 */
int OCSP_REQ_get_nonce(unsigned char *in, unsigned char *nonce, int sz) {

    char func[] = "OCSP_REQ_get_nonce()";


    /* OCSPRequest ::= SEQUENCE */
    if(in == NULL || *in != 0x30){ /* malformed data */
        sprintf(errmsg, "[%s] expected OCSPRequest ::= SEQUENCE", func);
        return -1;
    } 
    in = ASN1_next(in);

    /* TSBRequest ::= SEQUENCE */	
    if(in == NULL || *in != 0x30){ /* malformed data */
        sprintf(errmsg, "[%s] expected TSBRequest ::= SEQUENCE", func);
        return -1;
    } 

    
    while(*in) {        /* Search for Nonce in TSBRequest */  
        switch(*in) {
            case 0xa0:  /* version */
                    break;
            case 0xa1:  /* requestorName */
                    break;

            case 0x30:  /* requestList */
                    in = ASN1_next(in);
                    if(in == NULL) {
		      /* malformed data */
		      sprintf(errmsg, "malformed data");
		      return -1;
		    }
                    in = ASN1_skip(in);
                    break;

            case 0xa2:  /* requestExtensions */
                    return (getNonce(in, nonce, sz) == -1) ? -1 : 1;
            default:
	            sprintf(errmsg, "[%s] default case happen", func);
                    return -1;
        } /* end of switch */
    }
    return 1;
}


/*
 * getNonce()
 *
 * Get the Nonce value and set it to the area. 
 *
 * params in: pointer at 0xA0 in ASN.1
 *        out: area where Nonce is to be set
 *        sz: size of the out
 *
 * return -1: some error happened
 *        not -1: success
 */
int getNonce(unsigned char *in, char *out, int sz) {

    char func[] = "getNonce()";
    char oid[128];

    in = ASN1_next(in); /* SEQUENCE */
    in = ASN1_next(in); /* SEQUENCE */
    in = ASN1_next(in); /* OBJECT IDENTIFIER */

    /* check if it is Nonce */
    objid2str(in, oid, 128); 
    if (strcmp(OID_NONCE, oid) != 0) {
        sprintf(errmsg, "[%s] Not Nonce OID.", func);
	return -1;
    }

    in = ASN1_next(in);
    if(in[0] == 0x01) { /* check if it's BOOLEAN */
      in = ASN1_next(in);
    } else {
    }

    memset(out, 0x00, sz);
    memcpy(out, (in+2), *(in+1));
    nonceLen = *(in+1);

    return 1;
}


#ifdef DEBUG
/* dbgmsg
 */
void dbgmsg(char *func, char *msg) {
    printf("#DEBUG# %-16s : %s\n", func, msg);
}
#endif

/*
 * getValueNbytesFromLengthPointer()
 *
 *	Returns the length which is represented by bytes.
 *
 *	params in : pointer to ASN.1, which points at the beginning of
 *		    the length byte.
 */
int getValueNbytesFromLengthPointer(unsigned char *in) {
    int i, len, ret;

    ret = 0;
    if(0x80 & *in){  /* needs to see the following bytes for the length */
        len = 0x7f & *in;       /* a number of bytes for the length */
        for(i=0; i<len ;i++) {
            ret <<= 8;          /* sums up following values */
            ret |= in[i+1];
        }
        return(ret);

    } else {  /* the byte pointed by paramater is the length */
         return (0x7f & *in);
    }
}

/*
 * int derTagNbytes(unsigned char *derTLV)
 *
 *	returns the number of bytes of the tag for DER data 'derTLV'.
 *	Most of the case returns 1.
 *
 *	[...11111] specifies multi byte tag. (= lower 5 bits is 31.)
 *	[1.......] first byte of the tag value
 *	[1.......] second byte of the tag value
 *	    :
 *	[0.......] end of the tag value in which top most bit equals zero.
 */
int derTagNbytes(unsigned char *derTLV) {
  int count = 0;
  unsigned char *cp;

  if (*derTLV & 31) {
    cp = derTLV + 1;
    count = 1;
    while (*cp & 128) {		/* 128 = 10000000b */
      cp++;
      count++;
    }
    return count;
  } else {
    return 1;
  }
}

int derLengthNbytes(unsigned char *der) {
  if (*(der + 1) & 0x80) {
    return 1 + (*(der + 1) & 0x7f);
  } else {
    return 1;
  }
}


